package org.bitbucket.bradleysmithllc.etlunit.feature.log_assertion;

import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import org.apache.commons.io.FileUtils;
import org.bitbucket.bradleysmithllc.etlunit.*;
import org.bitbucket.bradleysmithllc.etlunit.context.VariableContext;
import org.bitbucket.bradleysmithllc.etlunit.feature.log_assertion.json.log_assertion.assert_log.AssertLogHandler;
import org.bitbucket.bradleysmithllc.etlunit.feature.log_assertion.json.log_assertion.assert_log.AssertLogRequest;
import org.bitbucket.bradleysmithllc.etlunit.feature.logging.LogFileManager;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestAnnotation;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestMethod;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestOperation;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestValueObject;

import javax.inject.Inject;
import java.io.IOException;
import java.util.List;
import java.util.regex.Pattern;

public class LogAssertionListener extends NullClassListener implements AssertLogHandler
{
	private LogAssertionRuntimeSupport logAssertionRuntimeSupport;
	private LogFileManager logFileManager;

	@Inject
	public void receiveLogAssertionRuntimeSupport(LogAssertionRuntimeSupport support)
	{
		logAssertionRuntimeSupport = support;
	}

	@Inject
	public void receiveLogFileManager(LogFileManager manager)
	{
		logFileManager = manager;
	}

	@Override
	public action_code assertLog(AssertLogRequest request, ETLTestOperation testOperation, ETLTestValueObject valueObject, VariableContext variableContext, ExecutionContext executionContext) throws TestAssertionFailure, TestExecutionError, TestWarning {
		assertLogImpl(request, testOperation.getTestMethod());

		return action_code.handled;
	}

	public void assertLogImpl(AssertLogRequest request, ETLTestMethod testMethod) throws TestAssertionFailure, TestExecutionError, TestWarning {
		String expectedLogFile = request.getExpectedLogFile();
		String expectedLogExpression = request.getExpectedLogExpression();

		if (expectedLogExpression == null && expectedLogFile == null)
		{
			throw new TestExecutionError("Log assertion requires exactly one of expected-log-expression or expected-log-file", LogAssertionConstants.ERR_LOG_ASSERTION_MISSING_EXPECTED_MESSAGE);
		}

		if (expectedLogExpression != null && expectedLogFile != null)
		{
			throw new TestExecutionError("Log assertion requires exactly one of expected-log-expression or expected-log-file", LogAssertionConstants.ERR_LOG_ASSERTION_DUPLICATE_EXPECTED_MESSAGE);
		}

		// resolve the text to compare
		String checkText = null;

		if (expectedLogExpression != null)
		{
			checkText = expectedLogExpression;
		}
		else
		{
			try {
				checkText = FileUtils.readFileToString(
					logAssertionRuntimeSupport.locateLogExpressionFileForCurrentTest(expectedLogFile)
				);
			} catch (IOException e) {
				throw new TestExecutionError("Could not read expected log file source [" + expectedLogFile + "]", LogAssertionConstants.ERR_LOG_ASSERTION_COULD_NOT_LOCATE_EXPECTED_LOG_FILE);
			}
		}

		LogFileManager.LogFile lf = getLogFile(request, testMethod, logFileManager);

		try {
			String logFileText = FileUtils.readFileToString(lf.getFile());

			AssertLogRequest.AssertionMode mode = request.getAssertionMode();

			String failureId = request.getFailureId();

			if (failureId == null)
			{
				failureId = LogAssertionConstants.FAIL_LOG_ASSERTION_LOG_FILE_DOES_NOT_MATCH;
			}

			// default mode is log-contains since it will be rare for an entire log to be predictable
			if (mode == null)
			{
				mode = AssertLogRequest.AssertionMode.LOG_CONTAINS;
			}

			switch (mode) {
				case LOG_EQUALS:
					if (!checkText.equals(logFileText))
					{
						throw new TestAssertionFailure("Log file [" + lf.getFile().getName() + "] does not equal expectation", failureId);
					}

					break;
				case LOG_MATCHES:
				{
					Pattern p = Pattern.compile(checkText);

					if (!p.matcher(logFileText).matches())
					{
						throw new TestAssertionFailure("Log file [" + lf.getFile().getName() + "] does not match expectation", failureId);
					}

					break;
				}
				case LOG_CONTAINS:
				{
					Pattern p = Pattern.compile(checkText);

					if (!p.matcher(logFileText).find())
					{
						throw new TestAssertionFailure("Log file [" + lf.getFile().getName() + "] does not contain expectation", failureId);
					}

					break;
				}
			}
		} catch (IOException e) {
			throw new TestExecutionError("Could not read actual log file source [" + expectedLogFile + "]", LogAssertionConstants.ERR_LOG_ASSERTION_COULD_READ_LOG_FILE);
		}
	}

	private LogFileManager.LogFile getLogFile(AssertLogRequest request, ETLTestMethod testMethod, LogFileManager logFileManager) throws TestExecutionError {
		// start off with a list of all logs created in this method
		List<LogFileManager.LogFile> logs = logFileManager.getLogFilesByETLTestMethod(testMethod);

		if (logs == null || logs.size() == 0)
		{
			throw new TestExecutionError("", LogAssertionConstants.ERR_LOG_ASSERTION_NO_LOGS);
		}

		// look for  file name pattern
		String namePattern = request.getLogNamePattern();
		Pattern nameP = null;

		if (namePattern != null)
		{
			nameP = Pattern.compile(namePattern);
		}

		// determine the traits we are looking for.  If there is no classifier we just go in order
		String classifier = request.getClassifier();

		// iterate backwards
		for (int i = logs.size() - 1; i >= 0; i--)
		{
			LogFileManager.LogFile logFile = logs.get(i);

			if (classifier == null || logFile.getClassifier().equals(classifier))
			{
				// double-check in case a name pattern was supplied
				if (nameP != null)
				{
					if (
						!nameP.matcher(logFile.getFile().getName()).find()
						&&
						!nameP.matcher(logFile.getOriginalPath().getName()).find()
					)
					{
						continue;
					}
				}

				// this one'll do
				return logFile;
			}
		}

		throw new TestExecutionError("Could not match log request to log file manager", LogAssertionConstants.ERR_LOG_ASSERTION_LOG_NOT_FOUND);
	}

	@Override
	public void end(ETLTestMethod mt, VariableContext context) throws TestAssertionFailure, TestExecutionError, TestWarning {
		// check for log assertion annotations.  Handle multiple even though it doesn't really make sense
		List<ETLTestAnnotation> annots = mt.getAnnotations("@LogAssertion");

		for (ETLTestAnnotation annot : annots)
		{
			processLogFileAssertion(mt, context, annot);
		}
	}

	private void processLogFileAssertion(ETLTestMethod mt, VariableContext context, ETLTestAnnotation annot) throws TestExecutionError, TestWarning, TestAssertionFailure {
		ETLTestValueObject obj = annot.getValue();

		ETLTestValueObject exp = obj.query("expected-log");
		ETLTestValueObject exps = obj.query("expected-logs");

		// sanity check
		if (exp != null && exps != null)
		{
			throw new TestExecutionError("Specify only one log option for @LogAssertion", LogAssertionConstants.ERR_LOG_ASSERTION_BAD_ANNOTATION);
		}
		else if (exp == null && exps == null)
		{
			throw new TestExecutionError("Specify a log option for @LogAssertion", LogAssertionConstants.ERR_LOG_ASSERTION_BAD_ANNOTATION_MISSING_OPTIONS);
		}

		if (exp != null)
		{
			processLogAssertion(mt, context, exp);
		}


		if (exps != null)
		{
			for (ETLTestValueObject la : exps.getValueAsList())
			{
				processLogAssertion(mt, context, la);
			}
		}
	}

	private void processLogAssertion(ETLTestMethod mt, VariableContext context, ETLTestValueObject exp) throws TestAssertionFailure, TestExecutionError, TestWarning {
		// convert into a json object so I can deserialize into the AssertLog
		JsonNode json = exp.getJsonNode();

		ObjectMapper om = new ObjectMapper();

		AssertLogRequest assertLogRequest = null;
		try {
			assertLogRequest = om.readValue(json.toString(), AssertLogRequest.class);
		} catch (IOException e) {
			throw new TestExecutionError("Could not deserialize annotation", TestConstants.ERR_IO_ERROR, e);
		}

		// pass on to the assertLog handler
		assertLogImpl(assertLogRequest, mt);
	}
}